/*
 *	Title:		Escavalon - Interactive Fiction System
 *	Version:	Beta
 *	Created by:	Paolo Copello
 *	Home page:	http://web.tiscali.it/fanelia
 *	
 *	Copyright (C) 2003 Paolo Copello
 *	
 *	******************************************
 *	This file is part of Escavalon
 *	******************************************
 *	
 *	Escavalon is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	any later version.
 *	
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *	
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *	
 */

#if !defined(FANEL_AVALON_ACTIONS_H__INCLUDED_)
#define FANEL_AVALON_ACTIONS_H__INCLUDED_

#include <string>
#include <list>
#include <stdexcept>

using namespace std;

#include "AvClass.h"

//class AvalonGame;

#define CALL_MEMBER_FN(object,ptrToMember)  ((object).*(ptrToMember)) 

///////////////////////////////////////////////////////////////

// non puo' essere istanziata

template <class T>
class AvAction
{
public:
	virtual ~AvAction() { }
	virtual bool Handle(T&, int nLocation, const vector<int>& command) = 0;
protected:
	AvAction() { }
};

///////////////////////////////////////////////////////////////

// non puo' essere istanziata

template <class T>
class AvActionWithList : public AvAction<T>
{
public:
	bool Handle(T&, int nLocation, const vector<int>& command);
	virtual void Execute(T&) = 0;
protected:
	AvActionWithList(int nLocation, const vector<int>& matchList);
	vector<int> _matchList;
	int _nLocation;
};



///////////////////////////////////////////////////////////////

// funziona solo con funzioni membro di T

template <class T>
class AvActionFunction : public AvActionWithList<T>
{
public:
	typedef void (T::*AV_MEMBER_HANDLER)();

	AvActionFunction(int nLocation, const vector<int>& matchList, AV_MEMBER_HANDLER pFunc );
	void Execute(T& game);
protected:
	AV_MEMBER_HANDLER _pFunc;
};

///////////////////////////////////////////////////////////////

// stampa una  stringa

template <class T>
class AvActionMessage : public AvActionWithList<T>
{
public:
	AvActionMessage(int nLocation, const vector<int>& matchList, const string& sMessage );
	void Execute(T& game);
protected:
	string _sMessage;
};


///////////////////////////////////////////////////////////////

// Timers Management
// funziona solo con funzioni membro di T

template <class T>
class AvTimerAction
{
public:
	typedef void (T::*AV_MEMBER_HANDLER)();
	AvTimerAction(int nTimerID, AV_MEMBER_HANDLER pFunc);
	void Execute(T& target);
private:
	int _nTimerID;
	AV_MEMBER_HANDLER _pFunc;
};

///////////////////////////////////////////////////////////////

template <class T>
class AvActionHandler
{
public:
	typedef void (T::*AV_MEMBER_HANDLER)();
	typedef list< AvAction<T>* > HandlersList;
	typedef Fanel::IntegerMap < AvTimerAction<T> > TimerHandlersList;

	AvActionHandler() { }
	~AvActionHandler();
	void AddAction(AvAction<T>* action);  // deve essere creato con NEW!

	// specific creation methods
	void AddAction(int nLoc, int nVerb, AV_MEMBER_HANDLER ptmf);
	void AddAction(int nLoc, int nVerb, int nTerm1, AV_MEMBER_HANDLER ptmf);
	void AddAction(int nLoc, int nVerb, int nTerm1, int nTerm2, AV_MEMBER_HANDLER ptmf);
	void AddAction(int nLoc, int nVerb, const string& sMessage);
	void AddAction(int nLoc, int nVerb, int nTerm1, const string& sMessage);
	void AddAction(int nLoc, int nVerb, int nTerm1, int nTerm2, const string& sMessage);

	void AddActionTimerElapsed(int nTimerID, AV_MEMBER_HANDLER ptmf);
//	void AddActionTimerTick(int nTimerID, AV_MEMBER_HANDLER ptmf);

	bool Handle(T& target, int nLocation, const vector<int>& command);

	bool HandleTimerElapsed(T& target, int nTimerID);
//	bool HandleTimerTick(T& target, int nTimerID);
private:
	HandlersList _handlers;
	TimerHandlersList _timerElapsed;
//	TimerHandlersList _timerTick;
};







///////////////////////////////////////////////////////////////
//	CLASS TEMPLATES IMPLEMENTATION
///////////////////////////////////////////////////////////////


/////////////////////////////////////
// AvActionWithList
/////////////////////////////////////

template <class T>
AvActionWithList<T>::AvActionWithList(int nLocation, const vector<int>& matchList)
: _matchList(matchList), _nLocation(nLocation)
{
}

template <class T>
bool AvActionWithList<T>::Handle(T& game, int nLocation, const vector<int>& command)
{
	// la location, merita un trattamento speciale
	if (_nLocation != AV_ANY_PLACE && _nLocation != nLocation)
		return false;
	// verifica i comandi
	if (command.size() != _matchList.size())
		return false;
	vector<int>::const_iterator itCmd = command.begin();
	vector<int>::const_iterator itMatch = _matchList.begin();
	for(;itCmd!=command.end(); ++itCmd, ++itMatch)
	{
		if( *itMatch != AV_ANY_TERM && *itMatch != *itCmd)
			return false;
	}
	// il comando e' stato riconosciuto
	Execute(game);
	return true;
}


/////////////////////////////////////
// AvActionFunction
/////////////////////////////////////

template <class T>
AvActionFunction<T>::AvActionFunction(int nLocation, const vector<int>& matchList, AV_MEMBER_HANDLER pFunc )
: AvActionWithList<T>(nLocation, matchList), _pFunc(pFunc)
{
	if(pFunc == NULL)
	{
		throw invalid_argument("NULL parameter passed to AvActionFunction constructor");
	}
}


template <class T>
void AvActionFunction<T>::Execute(T& game)
{
	CALL_MEMBER_FN(game, _pFunc)();
}

/////////////////////////////////////
// AvActionMessage
/////////////////////////////////////

template <class T>
AvActionMessage<T>::AvActionMessage(int nLocation, const vector<int>& matchList, const string& sMessage )
: AvActionWithList<T>(nLocation, matchList), _sMessage(sMessage)
{
	if(sMessage.empty())
	{
		throw invalid_argument("Empty string passed to AvActionMessage constructor");
	}
}

template <class T>
void AvActionMessage<T>::Execute(T& game)
{
	game.Io().Write(_sMessage);
	game.Io().WriteLine();
}



/////////////////////////////////////
// AvTimerAction
/////////////////////////////////////

template <class T>
AvTimerAction<T>::AvTimerAction(int nTimerID, AV_MEMBER_HANDLER pFunc)
: _nTimerID(nTimerID), _pFunc(pFunc)
{
}

template <class T>
void AvTimerAction<T>::Execute(T& target)
{
	CALL_MEMBER_FN(target, _pFunc)();
}



/////////////////////////////////////
// AvActionHandler
/////////////////////////////////////

template <class T>
AvActionHandler<T>::~AvActionHandler()
{
	HandlersList::iterator it = _handlers.begin();
	for(;it!=_handlers.end(); ++it)
	{
		delete (*it);
	}
}

template <class T>
void AvActionHandler<T>::AddAction(AvAction<T>* action)
{
	_handlers.push_back(action);
}

template <class T>
bool AvActionHandler<T>::Handle(T& target, int nLocation, const vector<int>& command)
{
	HandlersList::iterator it = _handlers.begin();
	for(;it!=_handlers.end(); ++it)
	{
		if( (*it)->Handle(target, nLocation, command))
			return true;
	}
	return false;
}

template <class T>
bool AvActionHandler<T>::HandleTimerElapsed(T& target, int nTimerID)
{
	if (_timerElapsed.Contains(nTimerID))
	{
		_timerElapsed[nTimerID].Execute(target);
		return true;
	}
	return false;
}

/*template <class T>
bool AvActionHandler<T>::HandleTimerTick(T& target, int nTimerID)
{
	if (_timerTick.Contains(nTimerID))
	{
		_timerTick[nTimerID].Execute(target);
		return true;
	}
	return false;
}*/


// creations functions

template <class T>
void AvActionHandler<T>::AddAction(int nLoc, int nVerb, void (T::*ptmf)())
{
	vector<int> v;
	v.push_back(nVerb);
	AddAction(new AvActionFunction<T>(nLoc, v, ptmf));
}

template <class T>
void AvActionHandler<T>::AddAction(int nLoc, int nVerb, int nTerm1, void (T::*ptmf)())
{
	vector<int> v;
	v.push_back(nVerb);
	v.push_back(nTerm1);
	AddAction(new AvActionFunction<T>(nLoc, v, ptmf));
}

template <class T>
void AvActionHandler<T>::AddAction(int nLoc, int nVerb, int nTerm1, int nTerm2, void (T::*ptmf)())
{
	vector<int> v;
	v.push_back(nVerb);
	v.push_back(nTerm1);
	v.push_back(nTerm2);
	AddAction(new AvActionFunction<T>(nLoc, v, ptmf));
}


template <class T>
void AvActionHandler<T>::AddAction(int nLoc, int nVerb, const string& sMessage)
{
	vector<int> v;
	v.push_back(nVerb);
	AddAction(new AvActionMessage<T>(nLoc, v, sMessage));
}

template <class T>
void AvActionHandler<T>::AddAction(int nLoc, int nVerb, int nTerm1, const string& sMessage)
{
	vector<int> v;
	v.push_back(nVerb);
	v.push_back(nTerm1);
	AddAction(new AvActionMessage<T>(nLoc, v, sMessage));
}

template <class T>
void AvActionHandler<T>::AddAction(int nLoc, int nVerb,  int nTerm1, int nTerm2,  const string& sMessage)
{
	vector<int> v;
	v.push_back(nVerb);
	v.push_back(nTerm1);
	v.push_back(nTerm2);
	AddAction(new AvActionMessage<T>(nLoc, v, sMessage));
}

template <class T>
void AvActionHandler<T>::AddActionTimerElapsed(int nTimerID, AV_MEMBER_HANDLER ptmf)
{
	_timerElapsed.Add(nTimerID, AvTimerAction<T>(nTimerID, ptmf));
}

/*template <class T>
void AvActionHandler<T>::AddActionTimerTick(int nTimerID, AV_MEMBER_HANDLER ptmf)
{
	_timerTick.Add(nTimerID, AvTimerAction<T>(nTimerID, ptmf));
}*/





#endif

